import uuid
import json
from mako.lookup import TemplateLookup
import iot_devices.host as host
import iot_devices.device as devices
from scullery import mqtt, messagebus
import logging
import time
import threading
import os
import weakref
from collections import OrderedDict
from weakref import WeakValueDictionary
lock = threading.Lock()


all_devs = weakref.WeakValueDictionary()


mqttlock = threading.Lock()

stopFlag = [0]


def scan():
    while 1:
        if stopFlag[0]:
            break
        time.sleep(10)
        if lock.acquire(timeout=1):
            try:
                for i in all_devs:
                    # If the last signal was very strong, we don't need to wait as long before considering
                    # it gone, because packet loss will be less
                    m = 3 if (all_devs[i].datapoints['rssi']
                              or -80) > -65 else 7

                    if all_devs[i].lastseen < time.monotonic() - (float(
                            all_devs[i].config.get('interval', 300) or 300) * m):
                        # This is how we mark it as not there
                        all_devs[i].set_data_point('rssi', -180)
            except Exception:
                logging.exception("RTL err")

            finally:
                lock.release()


t = threading.Thread(target=scan, name="RTL433Task", daemon=True)
t.start()


def stop():
    stopFlag[0] = 1


host.app_exit_register(stop)


templateGetter = TemplateLookup(os.path.dirname(__file__))

defaultSubclassCode = """
class CustomDeviceType(DeviceType):
    pass
"""


class RTL433Client(devices.Device):
    device_type = 'RTL433Client'
    readme = os.path.join(os.path.dirname(__file__), "README.md")
    defaultSubclassCode = defaultSubclassCode
    shortDescription = "This device lets you get data from a device using an RTL433 daemon and MQTT"

    def onConnectionChange(self, status):
        if status == "connected":
            self.set_data_point("mqttStatus", 1)
        else:
            self.set_data_point("mqttStatus", 0)

    def __init__(self, name, data, **kw):
        devices.Device.__init__(self, name, data, **kw)

        try:
            self.set_config_default('device.interval', '300')

            self.numeric_data_point(
                "rssi",
                min=-180,
                max=12,
                interval=float(self.config["device.interval"]),
                description="-75 if recetly seen, otherwise -180, we don't have real RSSI data",
                writable=False)

            self.set_alarm("No Signal", 'rssi', "value<-110")

            self.set_config_default('device.id', '')
            self.set_config_default('device.model', '')
            self.set_config_default('device.server', 'localhost')
            self.set_config_default('device.port', '1883')
            self.set_config_default('device.password', '')
            self.set_config_default('device.mqttTopic', 'home/rtl_433')

            # Pretend we have seen it to give the checker the right interval
            # before declaring it lost
            self.lastseen = time.monotonic()

            # This connection is actually  possibly shared
            # Scullery does the deduplication for us

            self.connection = mqtt.get_connection(
                self.config["device.server"],
                int(self.config["device.port"].strip() or 1883),
                password=self.config["device.password"].strip(),
                connection_id=str("RTL433Connection"))

            self.numeric_data_point("mqttStatus", writable=False)
            self.connection.subscribe_to_status(self.onConnectionChange)
            self.set_data_point("mqttStatus",
                                1 if self.connection.is_connected else 0)
            self.set_alarm("MQTT Lost", "mqttStatus",
                           "value < 0.5", auto_ack=True, trip_delay=10)

            topic = data.get("device.mqtttopic", "home/rtl_433")

            # We cannot use priority greater than info because these are unencrypted and untruusted and higher could make noise.

            def onBattery(t, m):
                m = float(m)
                if not 'battery' in self.datapoints:
                    self.numeric_data_point("battery",
                                            default=50,
                                            writable=False,
                                            unit="%")

                    # Always set before setting the alarm.
                    self.set_data_point("battery", m)

                    self.set_alarm(name="Low battery",
                                   datapoint="battery",
                                   expression="value < 15",
                                   priority="info",
                                   auto_ack=True)

                self.set_data_point("battery", m)

            def onWind(t, m):
                m = float(m)
                if not 'wind' in self.datapoints:
                    self.numeric_data_point("wind",
                                            unit="km/h",
                                            writable=False)

                    self.set_alarm(name="High Wind",
                                   datapoint="wind",
                                   expression="value > 35",
                                   priority="info")

                self.set_data_point("wind", m)

            def onTemp(t, m):
                m = float(m)
                if not 'temp' in self.datapoints:
                    self.numeric_data_point("temp",
                                            unit="degC",
                                            writable=False)
                    self.set_alarm(name="Freezing temperatures",
                                   datapoint="temp",
                                   expression="value < 2",
                                   priority="info")

                self.set_data_point("temp", m)

            def onHum(t, m):
                m = float(m)
                if not 'humidity' in self.datapoints:
                    self.numeric_data_point("humidity",
                                            unit="%",
                                            writable=False)
                    self.set_alarm(name="High humidity",
                                   datapoint="humidity",
                                   expression="value > 80",
                                   priority="info")
                    self.set_alarm(name="Low humidity",
                                   datapoint="humidity",
                                   expression="value < 20",
                                   priority="info",
                                   auto_ack=True)

                self.set_data_point("humidity", m)

            def onMoist(t, m):
                m = float(m)
                if not 'moisture' in self.datapoints:
                    self.numeric_data_point("moisture",
                                            unit="%",
                                            writable=False)
                self.set_data_point("moisture", m)

            def onPres(t, m):
                m = float(m)
                if not 'pressure' in self.datapoints:
                    self.numeric_data_point("pressure",
                                            unit="Pa",
                                            writable=False)
                self.set_data_point("pressure", m)

            def onWeight(t, m):
                m = float(m)
                if not 'weight' in self.datapoints:
                    self.numeric_data_point("weight", writable=False)
                self.set_data_point("weight", m)

            def onCommandCode(t, m):
                m = float(m)
                if not 'lastCommandCode' in self.datapoints:
                    self.object_data_point("lastCommandCode", writable=False)
                self.set_data_point("lastCommandCode", (m, time.time()))

            def onCommandName(t, m):
                m = float(m)
                if not 'lastCommandName' in self.datapoints:
                    self.object_data_point("lastCommandName", writable=False)
                self.set_data_point("lastCommandName", (m, time.time()))

            def onJSON(t, m):
                try:
                    m = json.loads(m)
                    self.print(m, "Saw packet on air")

                    # Going to do an ID match.
                    if 'device.id' in self.config and self.config['device.id']:
                        if not ('id' in m
                                and str(m['id']) == self.config['device.id']):
                            self.print(m, "Packet filter miss")
                            return

                    if 'device.model' in self.config and self.config['device.id']:
                        if not ('model' in m
                                and m['model'] == self.config['device.model']):
                            self.print(m, "Packet filter miss")
                            return

                    self.print(m, "Packet filter hit")

                    # No real RSSI, we just use -75 to mean good and -180 to mean bad.
                    self.set_data_point("rssi", -75)
                    self.set_alarm("Signal Lost", "rssi",
                                   "value < -98", auto_ack=True)

                    self.lastSeen = time.monotonic()

                    if 'humidity' in m:
                        onHum(0, m['humidity'])

                    if 'moisture' in m:
                        onMoist(0, m['moisture'])

                    if 'temperature_C' in m:
                        onTemp(0, m['temperature_C'])

                    if 'wind_avg_km_h' in m:
                        onWind(0, m['wind_avg_km_h'])

                    if 'pressure_kPa' in m:
                        onPres(0, m['pressure_kPa'] * 1000)

                    if 'pressure_hPa' in m:
                        onPres(0, m['pressure_hPa'] * 100)

                    # Keep a percent based API with randomly chosen high and low numbers
                    if 'battery_ok' in m:
                        onBattery(0, 100 if m['battery_ok'] else 5)

                    if 'cmd' in m:
                        onCommandCode(0, m['cmd'])

                    if 'button_id' in m:
                        onCommandCode(0, m['button_id'])

                    if 'button_name' in m:
                        onCommandName(0, m['button_name'])

                    if 'event' in m:
                        onCommandName(0, m['event'])

                    if 'code' in m:
                        onCommandCode(0, m['code'])
                except:
                    self.handle_exception()

            self.noGarbage = [onJSON]

            self.connection.subscribe(topic, onJSON, encoding="raw")

            all_devs[self.name] = self
        except Exception:
            self.handle_exception()

    def close(self):
        return super().close()
        self.connection.unsubscribe(self.noGarbage[0])

    def getManagementForm(self):
        return ""



from iot_devices import device


class ConfigurableAlarm(device.Device):
    """
        This class is just a simple wrapper for your framework's alarm functionality.
    """
    device_type = "ConfigurableAlarm"

    def __init__(self, name, data, **kw):
        device.Device.__init__(self, name, data, **kw)

        self.set_config_default("device.alarm_name", "alarm")
        self.set_config_default("device.priority", "warning")
        self.set_config_default("device.auto_acknowledge", "no")

        # When this is a 1, we alarm.
        self.numeric_data_point("trigger")

        self.set_alarm(self.config['device.alarm_name'],
                       "trigger",
                       expression="value>0.5",
                       priority=self.config['device.priority'],
                       auto_ack=self.config['device.auto_acknowledge'].lower()
                       in ('yes', 'true'))



from inspect import isawaitable
import os
import time
import logging

import colorzero

import iot_devices.device

logger = logging.Logger("plugins.zigbee2mqtt")


def hex_to_xy(hex):
    c = colorzero.Color.from_string(hex)
    x, y, z = c.xyz

    return (x / (x + y + z), y / (x + y + z), y)


def xy_to_hex(x, y, Y):
    Y = 1
    X = (x * Y) / y
    Z = ((1 - x - y) * Y) / y
    return colorzero.Color.from_xyz(X, Y, Z).html


class Zigbee2MQTT(iot_devices.device.Device):
    device_type = 'Zigbee2MQTT'

    description = "Connects to a Zigbee2MQTT gateway and makes the devices accessible via a tags API"

    def pair(self, t=120):
        # Enable pairing for 120 seconds
        self.connection.publish('zigbee2mqtt/bridge/request/permit_join', {
            "value": True,
            "time": t
        })

    def __init__(self, name, data, **kw):
        iot_devices.device.Device.__init__(self, name, data, **kw)
        self.devicesData = []

        self.knownDevices = {}

        self.nameToType = {}
        self.nameToTopic = {}
        self.nameToZBInfo = {}
        self.nameToHandler = {}

        self.set_config_default("device.mqtt_server", 'localhost')
        self.set_config_default("device.friendly_name", '__all__')

        try:
            from kaithem.src.scullery import mqtt

            # Ensure a new real connection.  This makes sure we get any retained messages.
            self.connection = mqtt.get_connection(
                self.config['device.mqtt_server'],
                connection_id=str(time.time()))

            self.connection.subscribe('zigbee2mqtt/bridge/devices',
                                      self.onDevices)
        except Exception:
            self.handleException()

    def on_data_point_change(self, tn, v, t, a):
        if not a == 'ZigBee':
            if tn in self.nameToZBInfo:
                type = self.nameToType[tn]
                topic = self.nameToTopic[tn]
                j = self.nameToZBInfo[tn]

                if type == "bool":
                    # Convert back to zigbee2mqtt's idea of true/false
                    self.connection.publish(topic + "/set", {
                        j['property']: (j['value_on'] if v else j['value_off'])
                    })

                elif type == 'color_xy':
                    # Convert back to zigbee2mqtt's idea of true/false
                    c = hex_to_xy(v)
                    self.connection.publish(
                        topic + "/set",
                        {j['property']: {
                            'x': c[0],
                            'y': c[1]
                        }})

                elif type == 'numeric' or type == 'string' or type == 'enum':
                    self.connection.publish(topic + "/set", {j['property']: v})

    def onDevices(self, t, v):
        self.devicesData = v

        d = {}
        for i in v:
            if not 'friendly_name' in i:
                continue
            d[i['friendly_name']] = True

            if 'definition' in i and i['definition']:
                if 'exposes' in i['definition']:

                    for f in i['definition']['exposes']:

                        # Normally the inner loop is just dealing with one expose
                        x = [f]

                        isWritable = (f.get("access", 2) & 2) > 0

                        isALight = False
                        # Sometimes there are multiple features
                        if f['type'] in ['switch', 'fan', 'climate']:
                            x = f['features']

                        if f['type'] in ['light']:
                            x = f['features']
                            isALight = True

                        for j in x:
                            if self.config['device.friendly_name'].strip() in (
                                    "*", "any", "__all__", ''):
                                tn = "node/" + i['friendly_name'] + '.' + j[
                                    'property']
                            else:
                                if not self.config[
                                        'device.friendly_name'].strip().lower(
                                ) == i['friendly_name'].lower():
                                    continue
                                tn = j['property']
                            zn = "zigbee2mqtt/" + i['friendly_name']

                            # Internally we represent all colors as the standard but kinda mediocre CSS strings,
                            # Since ZigBee seems to like to separate color and brightness, we will do the same thing here.
                            if j['name'] == 'color_xy' and isALight:
                                self.string_data_point(
                                    "node/" + i['friendly_name'] + "." +
                                    j['property'],
                                    subtype="color",
                                    writable=isWritable)

                                def f(t, v, tn=tn, j=j):
                                    if j['property'] in v:
                                        v = v[j['property']]
                                        v = xy_to_hex(v['x'], v['y'], 1)
                                        self.set_data_point(
                                            tn, v, annotation='ZigBee')

                                self.nameToTopic[tn] = zn
                                self.nameToZBInfo[tn] = j
                                self.nameToType[tn] = 'color_xy'

                                try:
                                    self.connection.unsubscribe(
                                        zn, self.nameToHandler[tn])
                                except:
                                    pass

                                self.connection.subscribe(zn, f)
                                self.nameToHandler[tn] = f

                            elif j['type'] == 'numeric':
                                self.numeric_data_point(
                                    tn,
                                    min=j.get("value_min", None),
                                    max=j.get("value_max", None),
                                    step=j.get("value_step", None),
                                    unit=j.get("unit", None).replace(
                                        "°", 'deg').replace("lqi", "%"),
                                    writable=isWritable)

                                if 'unit' in j:
                                    # Link quality low signal alarms
                                    if j['unit'] == 'lqi':
                                        self.set_alarm(name="LowSignal",
                                                       datapoint=tn,
                                                       expression="value < 8",
                                                       priority='info',
                                                       trip_delay=60)

                                    if j['name'] == 'battery':
                                        # Timestamp means wait for at least one actual data point
                                        self.set_alarm(name="LowBattery",
                                                       datapoint=tn,
                                                       expression="value < 33",
                                                       priority='warning')

                                def f(t, v, tn=tn, j=j):
                                    if j['property'] in v:
                                        self.set_data_point(
                                            tn,
                                            v[j['property']],
                                            annotation="ZigBee")

                                self.nameToTopic[tn] = zn
                                self.nameToZBInfo[tn] = j
                                self.nameToType[tn] = 'numeric'

                                try:
                                    self.connection.unsubscribe(
                                        zn, self.nameToHandler[tn])
                                except Exception:
                                    pass

                                self.connection.subscribe(zn, f)
                                self.nameToHandler[tn] = f

                            elif j['type'] == 'binary':
                                self.numeric_data_point(tn,
                                                        min=0,
                                                        max=1,
                                                        writable=isWritable)

                                if j['name'] == 'tamper':
                                    self.set_alarm(name="Tamper",
                                                   datapoint=tn,
                                                   expression="value",
                                                   priority='error')

                                if j['name'] == 'battery_low':
                                    self.set_alarm(name="BatteryLow",
                                                   datapoint=tn,
                                                   expression="value",
                                                   priority='warning')

                                if j['name'] == 'gas':
                                    self.set_alarm(name="GAS",
                                                   datapoint=tn,
                                                   expression="value",
                                                   priority='critical')

                                if j['name'] == 'smoke':
                                    self.set_alarm(name="SMOKE",
                                                   datapoint=tn,
                                                   expression="value",
                                                   priority='critical')

                                if j['name'] == 'water_leak':
                                    self.set_alarm(name="WATERLEAK",
                                                   datapoint=tn,
                                                   expression="value",
                                                   priority='critical')

                                if j['name'] == 'carbon_monoxide':
                                    self.set_alarm(name="CARBONMONOXIDE",
                                                   datapoint=tn,
                                                   expression="value",
                                                   priority='critical')

                                def f(t, v, tn=tn, j=j):
                                    if j['property'] in v:
                                        # Convert back to proper true/false
                                        v = v[j['property']] == j['value_on']
                                        self.set_data_point(
                                            tn,
                                            1 if v else 0,
                                            annotation='ZigBee')

                                self.nameToType[tn] = "bool"
                                self.nameToTopic[tn] = zn
                                self.nameToZBInfo[tn] = j

                                try:
                                    self.connection.unsubscribe(
                                        zn, self.nameToHandler[tn])
                                except Exception:
                                    pass

                                self.connection.subscribe(zn, f)
                                self.nameToHandler[tn] = f

                            # Todo: Tag points need to support enums
                            elif j['type'] in ['enum', 'text']:
                                self.string_data_point(tn, writable=isWritable)

                                def f(t, v, tn=tn, j=j):
                                    if j['property'] in v:
                                        self.set_data_point(
                                            tn,
                                            v[j['property']],
                                            annotation='ZigBee')

                                self.nameToType[tn] = j['type']
                                self.nameToTopic[tn] = zn
                                self.nameToZBInfo[tn] = j

                                try:
                                    self.connection.unsubscribe(
                                        zn, self.nameToHandler[tn])
                                except Exception:
                                    pass

                                self.connection.subscribe(zn, f)
                                self.nameToHandler[tn] = f

    def on_ui_message(self, v):
        # Note that this sends one wake request that lasts 30s only
        try:
            if v[0] == "enablePairing":
                self.pair()
                self.print(
                    "New ZigBee devices will be accepted for 120 seconds")

        except Exception:
            self.handleException()



from iot_devices import device
from iot_devices import host

import os
import getpass
import requests
import time
import threading
import json

from urllib.parse import quote


cache = "/dev/shm/wttr.in.cache/" + getpass.getuser()
if not os.path.exists(cache):
    os.makedirs(cache)

# if not os.path.exists(os.path.expanduser("~/wttr.in/cache")):
#     os.makedirs(os.path.expanduser("~/wttr.in/"),exist_ok=True)
#     os.link(os.path.expanduser("~/wttr.in/cache"), d)


def fetch(url, cachetime=1 * 3600):
    fn = quote(url).replace("/", "%2F")
    fn = os.path.join(cache, fn)
    if os.path.exists(fn):
        t = os.stat(fn).st_mtime
        if t > (time.time() - cachetime):
            with open(fn) as f:
                return f.read()

    d = requests.get(url)
    d.raise_for_status()

    with open(fn, 'w') as f:
        f.write(d.text)

    return d.text


def getWeather(place, cachetime=1 * 3600):
    return json.loads(fetch("https://wttr.in/" + place + "?format=j1", cachetime))



class WeatherClient(device.Device):
    device_type = "WeatherClient"

    def __init__(self, name, data, **kw):
        device.Device.__init__(self, name, data, **kw)

        if 'get_site_coordinates' in host.api:
            lat, lon = host.api['get_site_coordinates']()
        else:
            lat, lon = 90, 135

        if lat == None or lon == None:
            lat, lon = 90, 135

        location = str(lat) + "," + str(lon)

        self.shouldRun = True
        self.set_config_default("device.location", location)
        self.set_config_default("device.update_minutes", "180")
        self.config_properties['device.update_interval'] = {
            'description': "Values below 90 minutes are ignored",
            'unit': 'minute'
        }

        def worker():
            while self.shouldRun:
                try:
                    self.update()
                except Exception:
                    self.handle_exception()
                time.sleep(3600)

        self.thread = threading.Thread(
            target=worker, name="WeatherFetcher", daemon=True)

        # Push type data point set by the device
        self.numeric_data_point(
            "temperature", unit="degC", min=-100, max=85, writable=False)
        self.numeric_data_point("humidity", unit="%",
                                min=0, max=100, writable=False)
        self.numeric_data_point("wind", unit="KPH",
                                min=0, max=100, writable=False)
        self.numeric_data_point("pressure", unit="millibar",
                                min=100, max=1100, writable=False)
        self.numeric_data_point("uv_index",
                                min=0, max=20, writable=False)
        self.thread.start()

    def update(self):
        w = getWeather(self.config['device.location'], max(int(self.config['device.update_minutes'].strip()), 90))
        w2 = w['current_condition'][0]
        self.set_data_point('temperature', w2['temp_C'])
        self.set_data_point('humidity', w2['humidity'])
        self.set_data_point('pressure', w2['pressure'])
        self.set_data_point('uv_index', w2['uvIndex'])

    def close(self):
        self.shouldRun = False
        return super().close()


from iot_devices import device
import requests
import json
import time
import os
import threading
import copy
import traceback
from typing import Dict

import paho.mqtt.client as mqtt
from logging import getLogger
log = getLogger(__name__)

"""
Object representation for YoLink MQTT Client

"""

server_url = 'https://api.yosmart.com'
mqtt_url = 'api.yosmart.com'


readme = """
    Get the UAC ID and key in the advanced settings of the YoLink app.
    It will auto-discover all your devices.  Save this device again to update the list.
    Supports leak, temperature, door, and siren

    Note that kaithem's devices and tag points are named based on the discovered YoLink names.
    Changing the names in the YoLink app will break your automations and you will have to set them up again.

    Also, you can replace a lost sensor just by giving the replacement the same name, as the identity is linked to
    the name rather than the unique ID.

    This uses an UNENCRYPTED cloud API at the moment. Do NOT use from inside a network you don't trust.

    Also note that subdevices are created dynamically.  This means that your events and modules might be created before this device is finshed
    loading all your YoLink sensors.

    Since this uses a cloud API, it is very possible for the internet connection to fail. 
    Commands sent to devices while the network is out will be ignored.
"""


class RateLimiter():
    def __init__(self, count, duration) -> None:
        self.interval = duration / count
        self.count = count
        self.duration = duration

        self.lastRefill = time.monotonic()

    def limit(self):
        t = time.monotonic() - self.lastRefill
        self.lastRefill = time.monotonic()

        t = t / self.duration
        self.count = self.count + t

        if self.count < - self.count:
            raise RuntimeError("Too many attempts, too fast")

        if self.count < 1:
            time.sleep(self.interval)
            # Add a penalty for event trying to go too fast, so it eventually errors and someone notices
            self.count -= 0.5
        else:
            self.count -= 1


class YoLinkMQTTClient(object):

    def __init__(self, uid, key, homeid, parent, mqtt_port=8003, client_id=os.getpid()):

        self.topic = "yl-home/" + homeid + "/+/report"
        self.topic2 = "yl-home/" + homeid + "/+/response"

        self.mqtt_url = mqtt_url
        self.mqtt_port = int(mqtt_port)
        self.key = key
        self.parent = parent

        self.client = mqtt.Client(client_id=str(__name__ + str(client_id)),
                                  clean_session=True, userdata=None,
                                  protocol=mqtt.MQTTv311, transport="tcp")
        self.client.reconnect_delay_set(60, 30*60)
        self.client.on_connect = self.on_connect
        self.client.on_message = self.on_message

    def close(self):
        self.client.disconnect()
        self.client.loop_stop()

    def connect_to_broker(self):
        self.client.username_pw_set(username=self.key)
        self.client.on_message = self.on_message
        self.client.on_connect = self.on_connect

        self.client.connect(self.mqtt_url, self.mqtt_port, 10)
        self.client.loop_start()

    def on_message(self, client, userdata, msg):
        try:
            payload = json.loads(msg.payload.decode("utf-8"))

            if 'deviceId' in payload:
                deviceId = payload['deviceId']
            elif 'targetDevice' in payload and 'code' in payload:
                deviceId = payload['targetDevice']
            else:
                return

            if deviceId in self.parent.devices_by_id:
                self.parent.devices_by_id[deviceId].onData(payload)

        except Exception:
            self.parent.handle_error(traceback.format_exc())

    def on_connect(self, client, userdata, flags, rc):
        if (rc == 0):
            self.client.subscribe(self.topic)
            self.client.subscribe(self.topic2)

            self.parent.connected = True
            self.parent.set_data_point("connected", 1)
            self.parent.print("Connected to MQTT")

        else:
            self.parent.handle_error(traceback.format_exc())

    def on_disconnect(self, *a, **k):
        self.parent.set_data_point("connected", 0)


def listify(d):
    if not isinstance(d, list):
        d = [d]
    return d


# They name the exact same thing different names in reports and responses
# Handles:
# data :{state: normal}
# data :{state:{ state:normal}}
# state: {state: normal}
def get_from_state_or_data(d, p):
    if 'data' in d:
        x = d['data']
    elif 'state' in d:
        x = d['state']
    else:
        return None

    if 'state' in x:
        if isinstance(x['state'], dict):
            x = x['state']

    if p in x:
        return x[p]
    else:
        return None


class YoLinkDevice(device.Device):
    device_type = "YoLinkDevice"

    readme = readme

    has_battery = True

    def __init__(self, name: str, config: Dict[str, str], **kw):
        super().__init__(name, config, **kw)
        self.numeric_data_point("rssi", min=-120, max=8,
                                lo=-100, writable=False, default=-99)

        # Set a very long trip delay because of the slow updates
        self.set_alarm('Low Signal', 'rssi', 'value < -99', trip_delay=(3600*6)
                       if self.has_battery else (3600*4.5), auto_ack=True)

        if self.has_battery:
            self.numeric_data_point(
                "battery", min=0, max=100, lo=30, writable=False, auto_ack=True)
            self.set_alarm('Low Battery', 'battery', 'value < 30')

    def onData(self, data):

        if 'deviceId' in data:
            self.metadata['Device ID'] = data['deviceId']

        if 'data' in data:
            if 'loraInfo' in data['data']:
                s = data['data']['loraInfo']['signal']
                self.set_data_point('rssi', s)
                self.metadata['Gateway ID'] = data['data']['loraInfo']['gatewayId']

        if self.has_battery:
            battery = get_from_state_or_data(data, 'battery')

            if battery is not None:
                battery = battery * 25
                self.set_data_point('battery', battery)

        sl = get_from_state_or_data(data, 'soundLevel')

        if sl is not None:
            self.metadata['Sound Level Setting'] = sl

    def downlink(self, d):
        d.update({'token': self.token, 'targetDevice': self.deviceId})
        self.parent.sendDownlink(self.deviceId, d)

    def simpleMethod(self, m):
        self.downlink({'method': m})

    def refresh(self):
        pass


class YoLinkDoorSensor(YoLinkDevice):
    device_type = "YoLinkDoorSensor"

    def __init__(self, name: str, config: Dict[str, str], **kw):
        super().__init__(name, config, **kw)
        self.numeric_data_point("open", min=0, max=1,
                                subtype='bool', writable=False)

    def onData(self, data):
        YoLinkDevice.onData(self, data)
        o = get_from_state_or_data(data, 'state')
        if o is None:
            return

        open = 0 if (o == 'closed') else 1
        self.set_data_point('open', open)

    def refresh(self):
        self.simpleMethod("DoorSensor.getState")


class YoLinkOutlet(YoLinkDevice):
    device_type = "YoLinkOutlet"
    has_battery = False

    def __init__(self, name: str, config: Dict[str, str], **kw):
        super().__init__(name, config, **kw)
        self.numeric_data_point("switch", min=0, max=1,
                                subtype='bool', handler=self.setState)

        self.numeric_data_point("power", min=0, max=2500,
                                unit="W", writable=False)

        self.set_alarm("High Power Device", "power", "value > 600",
                       priority='info', auto_ack=True, trip_delay=10)
        self.set_alarm("Overload", "power", "value > 1500",
                       priority='critical', trip_delay=5)

    def onData(self, data):
        YoLinkDevice.onData(self, data)
        o = get_from_state_or_data(data, 'state')
        if o is None:
            return

        open = 0 if (o == 'closed') else 1
        self.set_data_point('switch', open)

    def setState(self, v, *a, **k):
        self.downlink({
            'method': "Outlet.setState",
            'params': {
                'state': 'open' if v else 'closed'
            }
        })

    def refresh(self):
        self.simpleMethod("Outlet.getState")


class YoLinkLeakSensor(YoLinkDevice):
    device_type = "YoLinkLeakSensor"

    def __init__(self, name: str, config: Dict[str, str], **kw):
        super().__init__(name, config, **kw)
        self.numeric_data_point("leak", min=0, max=1,
                                subtype='bool', writable=False)

        self.set_alarm('Water Detected', 'leak', 'value > 0',
                       priority='critical', auto_ack=True, trip_delay=5)

    def onData(self, data):
        YoLinkDevice.onData(self, data)
        o = get_from_state_or_data(data, 'state')
        if o is None:
            return

        leak = 0 if (o == 'normal') else 1
        self.set_data_point('leak', leak)

    def refresh(self):
        self.simpleMethod("LeakSensor.getState")


class YoLinkSiren(YoLinkDevice):
    device_type = "YoLinkSiren"

    def __init__(self, name: str, config: Dict[str, str], **kw):
        super().__init__(name, config, **kw)
        self.numeric_data_point(
            "on", min=0, max=1, subtype='bool', writable=False)

        self.numeric_data_point(
            "powered", min=0, max=1, lo=0, subtype='bool', writable=False)

        self.numeric_data_point("start", subtype='trigger',
                                description="Trigger a momentary siren", handler=self.doSiren)

        self.numeric_data_point("stop", subtype='trigger',
                                description="Stop an active siren", handler=self.stopSiren)

        self.set_alarm('PowerFail', 'powered', 'value < 1',
                       priority='error')

        self.set_alarm('Siren "+name+" is on', 'on', 'value > 0',
                       priority='warning', auto_ack=True)

    def refresh(self):
        self.simpleMethod("Siren.getState")

    def doSiren(self, v, *a, **k):
        self.downlink({
            'method': "Siren.setState",
            'params': {
                'state':
                {'alarm': True}
            }
        })

    def stopSiren(self, v, *a, **k):
        self.downlink({
            'method': "Siren.setState",
            'params': {
                'state':
                {'alarm': False}
            }
        })

    def onData(self, data):
        YoLinkDevice.onData(self, data)
        o = get_from_state_or_data(data, 'state')
        if o is None:
            return

        on = 0 if (o == 'normal') else 1
        self.set_data_point('on', on)

        o = get_from_state_or_data(data, 'powerSupply')
        if o is None:
            return

        usb = 1 if (data['data']['powerSupply'] == 'usb') else 0
        self.set_data_point('powered', usb)


class YoLinkTemperatureSensor(YoLinkDevice):
    device_type = "YoLinkTemperatureSensor"

    def __init__(self, name: str, config: Dict[str, str], **kw):
        super().__init__(name, config, **kw)
        self.numeric_data_point("temperature", min=-40,
                                max=100, unit='degC', writable=False)

        self.numeric_data_point(
            "humidity", min=0, max=100, unit='%', writable=False)

        self.set_alarm("Extreme high temperature", "temperature",
                       "value > 85", priority="critical")
        self.set_alarm("Very high humidity", "humidity",
                       "value > 90", priority="info", auto_ack=True)

    def onData(self, data):
        YoLinkDevice.onData(self, data)

        t = get_from_state_or_data(data, 'temperature')
        if t is None:
            return

        h = get_from_state_or_data(data, 'humidity')
        if h is None:
            return

        self.set_data_point('temperature', t)
        self.set_data_point('humidity', t)

    def refresh(self):
        self.simpleMethod("THSensor.getState")


deviceTypes = {
    "YolinkDevice": YoLinkDevice,
    "DoorSensor": YoLinkDoorSensor,
    "LeakSensor": YoLinkLeakSensor,
    "THSensor": YoLinkTemperatureSensor,
    "Siren": YoLinkSiren,
    "Outlet": YoLinkOutlet
}

connectRateLimit = RateLimiter(5, 5 * 60)


class YoLinkService(device.Device):
    device_type = "YoLinkService"

    def makeRequest(self, r):
        r = copy.deepcopy(r)
        r['time'] = int(time.time())

        return json.loads(requests.post(server_url + '/open/yolink/v2/api', headers={
            "Content-Type": "application/json",
            "Authorization": "Bearer " + self.token,
        }, data=json.dumps(r), timeout=5).text)['data']

    def close(self):
        try:
            self.client.close()
        except Exception:
            log.exception("Error closing")
        return super().close()

    def sendDownlink(self, device, data):
        self.dowlinkRateLimit.limit()
        try:
            data['time'] = int(time.time())

            self.client.client.publish(
                "yl-home" + "/" + self.homeId + "/" + device + "/request", json.dumps(data))

        except Exception:
            self.handle_error(traceback.format_exc())

    def initialConnection(self):
        self.token = requests.post(server_url + '/open/yolink/token', params={
            'grant_type': 'client_credentials',
            'client_id': self.config['device.user_id'],
            'client_secret': self.config['device.key'],
        }, timeout=5).text

        r = {
            "method": "Home.getDeviceList",
        }
        self.token = json.loads(self.token)['access_token']
        devList = listify(self.makeRequest(r)['devices'])

        r = {
            "method": "Home.getGeneralInfo",
        }

        gInfo = self.makeRequest(r)

        self.homeId = listify(gInfo)[0]['id']

        c = YoLinkMQTTClient(
            self.config['device.user_id'], self.token, self.homeId, self)
        self.client = c

        self.connected = False
        c.connect_to_broker()

        t = time.monotonic()

        while not self.connected:
            if time.monotonic() - t > 5:
                break

        for i in devList:
            t = i['type']
            d = {

            }

            if t in deviceTypes:
                d = self.create_subdevice(
                    deviceTypes[t], i['name'], d)

                d.deviceId = i['deviceId']
                d.token = i['token']
                d.parent = self

                self.devices_by_id[i['deviceId']] = d
        for i in self.devices_by_id:
            try:
                self.devices_by_id[i].refresh()
            except Exception:
                self.handle_error(traceback.format_exc())

        self.initialConnectionDone = True

    def retryLoop(self):
        while self.shouldRun:
            try:
                if self.initialConnectionDone:
                    time.sleep(60*10)
                    continue

                time.sleep(60*10)

                with self.connectLock:
                    if not self.initialConnectionDone:
                        connectRateLimit.limit()
                        self.initialConnection()
            except Exception:
                self.handle_error(traceback.format_exc())

    def __init__(self, name, data, **kw):
        device.Device.__init__(self, name, data, **kw)
        self.shouldRun = False
        try:
            self.set_config_default("device.user_id", "")
            self.set_config_default("device.key", "")

            self.numeric_data_point(
                "connected", subtype="bool", writable=False)
            self.set_alarm("Dis_connected from YoLink API", 'connected',
                           "value < 1", priority="warning", trip_delay=25)

            self.dowlinkRateLimit = RateLimiter(60, 60 * 10)

            self.devices_by_id = {}

            self.initialConnectionDone = False

            self.connectLock = threading.RLock()

            if self.config['device.key']:
                with self.connectLock:
                    if not self.initialConnectionDone:
                        connectRateLimit.limit()
                        try:
                            self.initialConnection()
                        except Exception:
                            self.handle_error(traceback.format_exc())
                            t = threading.Thread(
                                target=self.retryLoop, name="YoLinkConnectionRetry")
                            t.start()

        except Exception:
            self.handle_error(traceback.format_exc())

    @classmethod
    def discover_devices(cls, config={}, current_device=None, intent=None, **kw):
        return {}



from setuptools import setup, find_packages

setup(
    name='iot_devices',
    version='0.1.15',
    author="Daniel Dunn",
    author_email="dannydunn@eternityforest.com",
    packages=find_packages(),
    package_data={'': ["*.json"]},
    license='MIT',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    url="https://github.com/EternityForest/iot_devices",
    entry_points={
        'console_scripts': [
            'tui-dash = iot_devices.tui_dash:main',
        ],
    },
    install_requires=[
        "paho.mqtt",
        "urwid",
        "scullery",
        "yeelight",
        "colorzero"
    ]
)
